import math
import random
import logging
import json
import functools
import asyncio
import aiohttp
import secrets
from asyncinit import asyncinit

from hailtop.config import get_deploy_config
from hailtop.auth import async_get_userinfo, service_auth_headers
from hailtop.utils import bounded_gather, request_retry_transient_errors, tqdm, TQDM_DEFAULT_DISABLE

from .globals import tasks, complete_states

log = logging.getLogger('batch_client.aioclient')


class Job:
    @staticmethod
    def _get_error(job_status, task):
        status = job_status.get('status')
        if not status:
            return None

        # don't return status error

        container_statuses = status.get('container_statuses')
        if not container_statuses:
            return None

        container_status = container_statuses.get(task)
        if not container_status:
            return None

        error = container_status.get('error')
        if error:
            return error

        docker_container_status = container_status.get('container_status')
        if not docker_container_status:
            return None

        return docker_container_status.get('error')

    @staticmethod
    def _get_out_of_memory(job_status, task):
        status = job_status.get('status')
        if not status:
            return None

        container_statuses = status.get('container_statuses')
        if not container_statuses:
            return None

        container_status = container_statuses.get(task)
        if not container_status:
            return None

        docker_container_status = container_status.get('container_status')
        if not docker_container_status:
            return None

        return docker_container_status['out_of_memory']

    @staticmethod
    def _get_container_status_exit_code(container_status):
        docker_container_status = container_status.get('container_status')
        if not docker_container_status:
            return None

        return docker_container_status.get('exit_code')

    @staticmethod
    def _get_exit_code(job_status, task):
        status = job_status.get('status')
        if not status:
            return None

        container_statuses = status.get('container_statuses')
        if not container_statuses:
            return None

        container_status = container_statuses.get(task)
        if not container_status:
            return None

        return Job._get_container_status_exit_code(container_status)

    @staticmethod
    def _get_exit_codes(job_status):
        status = job_status.get('status')
        if not status:
            return None

        error = status.get('error')
        if error is not None:
            return None

        container_statuses = status.get('container_statuses')
        if not container_statuses:
            return None

        return {
            task: Job._get_container_status_exit_code(container_status)
            for task, container_status in container_statuses.items()
        }

    @staticmethod
    def exit_code(job_status):
        exit_codes = Job._get_exit_codes(job_status)
        if exit_codes is None:
            return None

        exit_codes = [
            exit_codes[task]
            for task in tasks
            if task in exit_codes
        ]

        i = 0
        while i < len(exit_codes):
            ec = exit_codes[i]
            if ec is None:
                return None
            if ec > 0:
                return ec
            i += 1
        return 0

    @staticmethod
    def total_duration_msecs(job_status):
        status = job_status.get('status')
        if not status:
            return None

        container_statuses = status.get('container_statuses')
        if not container_statuses:
            return None

        def _get_duration(container_status):
            if not container_status:
                return None

            timing = container_status.get('timing')
            if not timing:
                return None

            runtime = timing.get('runtime')
            if not runtime:
                return None

            return runtime.get('duration')

        durations = [
            _get_duration(container_status)
            for task, container_status in container_statuses.items()
        ]

        if any(d is None for d in durations):
            return None
        return sum(durations)

    @staticmethod
    def unsubmitted_job(batch_builder, job_id):
        assert isinstance(batch_builder, BatchBuilder)
        _job = UnsubmittedJob(batch_builder, job_id)
        return Job(_job)

    @staticmethod
    def submitted_job(batch, job_id, _status=None):
        assert isinstance(batch, Batch)
        _job = SubmittedJob(batch, job_id, _status)
        return Job(_job)

    def __init__(self, job):
        self._job = job

    @property
    def batch_id(self):
        return self._job.batch_id

    @property
    def job_id(self):
        return self._job.job_id

    @property
    def id(self):
        return self._job.id

    async def attributes(self):
        return await self._job.attributes()

    async def is_complete(self):
        return await self._job.is_complete()

    async def status(self):
        return await self._job.status()

    @property
    def _status(self):
        return self._job._status

    async def wait(self):
        return await self._job.wait()

    async def log(self):
        return await self._job.log()


class UnsubmittedJob:
    def _submit(self, batch):
        return SubmittedJob(batch, self._job_id)

    def __init__(self, batch_builder, job_id):
        self._batch_builder = batch_builder
        self._job_id = job_id

    @property
    def batch_id(self):
        raise ValueError("cannot get the batch_id of an unsubmitted job")

    @property
    def job_id(self):
        raise ValueError("cannot get the job_id of an unsubmitted job")

    @property
    def id(self):
        raise ValueError("cannot get the id of an unsubmitted job")

    async def attributes(self):
        raise ValueError("cannot get the attributes of an unsubmitted job")

    async def is_complete(self):
        raise ValueError("cannot determine if an unsubmitted job is complete")

    async def status(self):
        raise ValueError("cannot get the status of an unsubmitted job")

    @property
    def _status(self):
        raise ValueError("cannot get the _status of an unsubmitted job")

    async def wait(self):
        raise ValueError("cannot wait on an unsubmitted job")

    async def log(self):
        raise ValueError("cannot get the log of an unsubmitted job")


class SubmittedJob:
    def __init__(self, batch, job_id, _status=None):
        self._batch = batch
        self.batch_id = batch.id
        self.job_id = job_id
        self.id = (self.batch_id, self.job_id)
        self._status = _status

    async def attributes(self):
        if not self._status:
            await self.status()
        return self._status['attributes']

    async def is_complete(self):
        if self._status:
            state = self._status['state']
            if state in complete_states:
                return True
        await self.status()
        state = self._status['state']
        return state in complete_states

    async def status(self):
        resp = await self._batch._client._get(f'/api/v1alpha/batches/{self.batch_id}/jobs/{self.job_id}')
        self._status = await resp.json()
        return self._status

    async def wait(self):
        i = 0
        while True:
            if await self.is_complete():
                return self._status
            j = random.randrange(math.floor(1.1 ** i))
            await asyncio.sleep(0.100 * j)
            # max 44.5s
            if i < 64:
                i = i + 1

    async def log(self):
        resp = await self._batch._client._get(f'/api/v1alpha/batches/{self.batch_id}/jobs/{self.job_id}/log')
        return await resp.json()


class Batch:
    def __init__(self, client, id, attributes, n_jobs):
        self._client = client
        self.id = id
        self.attributes = attributes
        self.n_jobs = n_jobs

    async def cancel(self):
        await self._client._patch(f'/api/v1alpha/batches/{self.id}/cancel')

    async def jobs(self, q=None):
        last_job_id = None
        while True:
            params = {}
            if q is not None:
                params['q'] = q
            if last_job_id is not None:
                params['last_job_id'] = last_job_id
            resp = await self._client._get(f'/api/v1alpha/batches/{self.id}/jobs', params=params)
            body = await resp.json()
            for job in body['jobs']:
                yield job
            last_job_id = body.get('last_job_id')
            if last_job_id is None:
                break

    async def status(self):
        resp = await self._client._get(f'/api/v1alpha/batches/{self.id}')
        return await resp.json()

    async def wait(self, *, disable_progress_bar=TQDM_DEFAULT_DISABLE):
        i = 0
        with tqdm(total=self.n_jobs,
                  disable=disable_progress_bar,
                  desc='completed jobs') as pbar:
            while True:
                status = await self.status()
                pbar.update(status['n_completed'] - pbar.n)
                if status['complete']:
                    return status
                j = random.randrange(math.floor(1.1 ** i))
                await asyncio.sleep(0.100 * j)
                # max 44.5s
                if i < 64:
                    i = i + 1

    async def delete(self):
        await self._client._delete(f'/api/v1alpha/batches/{self.id}')


class BatchBuilder:
    def __init__(self, client, attributes, callback):
        self._client = client
        self._job_idx = 0
        self._job_specs = []
        self._jobs = []
        self._submitted = False
        self.attributes = attributes
        self.callback = callback

    def create_job(self, image, command, env=None, mount_docker_socket=False,
                   port=None, resources=None, secrets=None,
                   service_account=None, attributes=None, parents=None,
                   input_files=None, output_files=None, always_run=False, pvc_size=None,
                   timeout=None):
        if self._submitted:
            raise ValueError("cannot create a job in an already submitted batch")

        self._job_idx += 1

        if parents is None:
            parents = []

        parent_ids = []
        foreign_batches = []
        invalid_job_ids = []
        for parent in parents:
            job = parent._job
            if isinstance(job, UnsubmittedJob):
                if job._batch_builder != self:
                    foreign_batches.append(job)
                elif not 0 < job._job_id < self._job_idx:
                    invalid_job_ids.append(job)
                else:
                    parent_ids.append(job._job_id)
            else:
                foreign_batches.append(job)

        error_msg = []
        if len(foreign_batches) != 0:
            error_msg.append('Found {} parents from another batch:\n{}'.format(str(len(foreign_batches)),
                                                                               "\n".join([str(j) for j in foreign_batches])))
        if len(invalid_job_ids) != 0:
            error_msg.append('Found {} parents with invalid job ids:\n{}'.format(str(len(invalid_job_ids)),
                                                                                 "\n".join([str(j) for j in invalid_job_ids])))
        if error_msg:
            raise ValueError("\n".join(error_msg))

        job_spec = {
            'always_run': always_run,
            'command': command,
            'image': image,
            'job_id': self._job_idx,
            'mount_docker_socket': mount_docker_socket,
            'parent_ids': parent_ids
        }

        if env:
            job_spec['env'] = [{'name': k, 'value': v} for (k, v) in env.items()]
        if port is not None:
            job_spec['port'] = port
        if resources:
            job_spec['resources'] = resources
        if secrets:
            job_spec['secrets'] = secrets
        if service_account:
            job_spec['service_account'] = service_account
        if timeout:
            job_spec['timeout'] = timeout

        if attributes:
            job_spec['attributes'] = attributes
        if input_files:
            job_spec['input_files'] = [{"from": src, "to": dst} for (src, dst) in input_files]
        if output_files:
            job_spec['output_files'] = [{"from": src, "to": dst} for (src, dst) in output_files]
        if pvc_size:
            job_spec['pvc_size'] = pvc_size

        self._job_specs.append(job_spec)

        j = Job.unsubmitted_job(self, self._job_idx)
        self._jobs.append(j)
        return j

    async def _submit_jobs(self, batch_id, byte_job_specs, n_jobs, pbar):
        assert len(byte_job_specs) > 0, byte_job_specs

        b = bytearray()
        b.append(ord('['))

        i = 0
        while i < len(byte_job_specs):
            spec = byte_job_specs[i]
            if i > 0:
                b.append(ord(','))
            b.extend(spec)
            i += 1

        b.append(ord(']'))

        await self._client._post(
            f'/api/v1alpha/batches/{batch_id}/jobs/create',
            data=aiohttp.BytesPayload(
                b, content_type='application/json', encoding='utf-8'))
        pbar.update(n_jobs)

    async def _create(self, batch_token=None):
        batch_token = batch_token or secrets.token_urlsafe(32)
        n_jobs = len(self._job_specs)
        batch_spec = {'billing_project': self._client.billing_project,
                      'n_jobs': n_jobs,
                      'token': batch_token}
        if self.attributes:
            batch_spec['attributes'] = self.attributes
        if self.callback:
            batch_spec['callback'] = self.callback
        batch_json = await (await self._client._post('/api/v1alpha/batches/create',
                                                     json=batch_spec)).json()
        return Batch(self._client, batch_json['id'], self.attributes, n_jobs)

    MAX_BUNCH_BYTESIZE = 1024 * 1024
    MAX_BUNCH_SIZE = 1024

    async def submit(self,
                     max_bunch_bytesize=MAX_BUNCH_BYTESIZE,
                     max_bunch_size=MAX_BUNCH_SIZE,
                     disable_progress_bar=TQDM_DEFAULT_DISABLE):
        assert max_bunch_bytesize > 0
        assert max_bunch_size > 0
        if self._submitted:
            raise ValueError("cannot submit an already submitted batch")
        batch = await self._create()
        id = batch.id
        log.info(f'created batch {id}')
        byte_job_specs = [json.dumps(job_spec).encode('utf-8')
                          for job_spec in self._job_specs]
        byte_job_specs_bunches = []
        bunch_sizes = []
        bunch = []
        bunch_n_bytes = 0
        bunch_n_jobs = 0
        for spec in byte_job_specs:
            n_bytes = len(spec)
            assert n_bytes < max_bunch_bytesize, (
                f'every job spec must be less than max_bunch_bytesize,'
                f' { max_bunch_bytesize }B, but {spec} is larger')
            if bunch_n_bytes + n_bytes < max_bunch_bytesize and len(bunch) < max_bunch_size:
                bunch.append(spec)
                bunch_n_bytes += n_bytes
                bunch_n_jobs += 1
            else:
                byte_job_specs_bunches.append(bunch)
                bunch_sizes.append(bunch_n_jobs)
                bunch = [spec]
                bunch_n_bytes = n_bytes
                bunch_n_jobs = 1
        if bunch:
            byte_job_specs_bunches.append(bunch)
            bunch_sizes.append(bunch_n_jobs)

        with tqdm(total=len(self._job_specs),
                  disable=disable_progress_bar,
                  desc='jobs submitted to queue') as pbar:
            await bounded_gather(
                *[functools.partial(self._submit_jobs, id, bunch, size, pbar)
                  for bunch, size in zip(byte_job_specs_bunches, bunch_sizes)],
                parallelism=6)

        await self._client._patch(f'/api/v1alpha/batches/{id}/close')
        log.info(f'closed batch {id}')

        for j in self._jobs:
            j._job = j._job._submit(batch)

        self._job_specs = []
        self._jobs = []
        self._job_idx = 0

        self._submitted = True
        return batch


@asyncinit
class BatchClient:
    async def __init__(self, billing_project, deploy_config=None, session=None,
                       headers=None, _token=None):
        self.billing_project = billing_project

        if not deploy_config:
            deploy_config = get_deploy_config()

        self.url = deploy_config.base_url('batch')

        if session is None:
            session = aiohttp.ClientSession(raise_for_status=True,
                                            timeout=aiohttp.ClientTimeout(total=60))
        self._session = session

        userinfo = await async_get_userinfo(deploy_config)
        self.bucket = userinfo['bucket_name']

        h = {}
        if headers:
            h.update(headers)
        if _token:
            h['Authorization'] = f'Bearer {_token}'
        else:
            h.update(service_auth_headers(deploy_config, 'batch'))
        self._headers = h

    async def _get(self, path, params=None):
        return await request_retry_transient_errors(
            self._session, 'GET',
            self.url + path, params=params, headers=self._headers)

    async def _post(self, path, data=None, json=None):
        return await request_retry_transient_errors(
            self._session, 'POST',
            self.url + path, data=data, json=json, headers=self._headers)

    async def _patch(self, path):
        return await request_retry_transient_errors(
            self._session, 'PATCH',
            self.url + path, headers=self._headers)

    async def _delete(self, path):
        return await request_retry_transient_errors(
            self._session, 'DELETE',
            self.url + path, headers=self._headers)

    async def list_batches(self, q=None):
        last_batch_id = None
        while True:
            params = {}
            if q is not None:
                params['q'] = q
            if last_batch_id is not None:
                params['last_batch_id'] = last_batch_id

            resp = await self._get('/api/v1alpha/batches', params=params)
            body = await resp.json()

            for batch in body['batches']:
                yield Batch(self, batch['id'], attributes=batch.get('attributes'), n_jobs=int(batch['n_jobs']))
            last_batch_id = body.get('last_batch_id')
            if last_batch_id is None:
                break

    async def get_job(self, batch_id, job_id):
        b = await self.get_batch(batch_id)
        j_resp = await self._get(f'/api/v1alpha/batches/{batch_id}/jobs/{job_id}')
        j = await j_resp.json()
        return Job.submitted_job(
            b,
            j['job_id'],
            _status=j)

    async def get_job_log(self, batch_id, job_id):
        resp = await self._get(
            f'/api/v1alpha/batches/{batch_id}/jobs/{job_id}/log')
        return await resp.json()

    async def get_batch(self, id):
        b_resp = await self._get(f'/api/v1alpha/batches/{id}')
        b = await b_resp.json()
        return Batch(self,
                     b['id'],
                     attributes=b.get('attributes'),
                     n_jobs=int(b['n_jobs']))

    def create_batch(self, attributes=None, callback=None):
        return BatchBuilder(self, attributes, callback)

    async def close(self):
        await self._session.close()
        self._session = None



from hail import ir
import abc
from typing import Sequence, MutableSequence, List, Set, Dict, Optional
from collections import namedtuple


class Renderable(object):
    @abc.abstractmethod
    def render_head(self, r: 'Renderer') -> str:
        ...

    @abc.abstractmethod
    def render_tail(self, r: 'Renderer') -> str:
        ...

    @abc.abstractmethod
    def render_children(self, r: 'Renderer') -> Sequence['Renderable']:
        ...


class RenderableStr(Renderable):
    def __init__(self, s: str):
        self.s = s

    def render_head(self, r: 'Renderer') -> str:
        return self.s

    @abc.abstractmethod
    def render_tail(self, r: 'Renderer') -> str:
        return ''

    @abc.abstractmethod
    def render_children(self, r: 'Renderer') -> Sequence['Renderable']:
        return []


class ParensRenderer(Renderable):
    def __init__(self, rs: Sequence['Renderable']):
        self.rs = rs

    def render_head(self, r: 'Renderer') -> str:
        return '('

    @abc.abstractmethod
    def render_tail(self, r: 'Renderer') -> str:
        return ')'

    @abc.abstractmethod
    def render_children(self, r: 'Renderer') -> Sequence['Renderable']:
        return self.rs


class RenderableQueue(object):
    def __init__(self, elements: Sequence[Renderable], tail: str):
        self._elements = elements
        self._elements_len = len(elements)
        self.tail = tail
        self._idx = 0

    def exhausted(self):
        return self._elements_len == self._idx

    def pop(self) -> Renderable:
        idx = self._idx
        self._idx += 1
        return self._elements[idx]


class RQStack(object):
    def __init__(self):
        self._list: List[RenderableQueue] = []
        self._idx = -1

    def push(self, x: RenderableQueue):
        self._idx += 1
        self._list.append(x)

    def peek(self) -> RenderableQueue:
        return self._list[self._idx]

    def pop(self) -> RenderableQueue:
        self._idx -= 1
        return self._list.pop()

    def non_empty(self) -> bool:
        return self._idx >= 0

    def is_empty(self) -> bool:
        return self._idx < 0


class Renderer:
    @abc.abstractmethod
    def add_jir(self, jir):
        pass


class PlainRenderer(Renderer):
    def __init__(self, stop_at_jir=False):
        self.stop_at_jir = stop_at_jir
        self.count = 0
        self.jirs = {}

    def add_jir(self, jir):
        jir_id = f'm{self.count}'
        self.count += 1
        self.jirs[jir_id] = jir
        return jir_id

    def __call__(self, x: 'Renderable'):
        stack = RQStack()
        builder = []

        while x is not None or stack.non_empty():
            if x is not None:
                # TODO: it would be nice to put the JavaIR logic in BaseIR somewhere but this isn't trivial
                if self.stop_at_jir and hasattr(x, '_jir'):
                    jir_id = self.add_jir(x._jir)
                    if isinstance(x, ir.MatrixIR):
                        builder.append(f'(JavaMatrix {jir_id})')
                    elif isinstance(x, ir.TableIR):
                        builder.append(f'(JavaTable {jir_id})')
                    elif isinstance(x, ir.BlockMatrixIR):
                        builder.append(f'(JavaBlockMatrix {jir_id})')
                    else:
                        assert isinstance(x, ir.IR)
                        builder.append(f'(JavaIR {jir_id})')
                else:
                    head = x.render_head(self)
                    if head != '':
                        builder.append(x.render_head(self))
                    stack.push(RenderableQueue(x.render_children(self), x.render_tail(self)))
                x = None
            else:
                top = stack.peek()
                if top.exhausted():
                    stack.pop()
                    builder.append(top.tail)
                else:
                    builder.append(' ')
                    x = top.pop()

        return ''.join(builder)


Vars = Dict[str, int]
Context = (Vars, Vars, Vars)


BindingSite = namedtuple(
    'BindingSite',
    'depth lifted_lets agg_lifted_lets scan_lifted_lets')


class CSERenderer(Renderer):
    def __init__(self, stop_at_jir=False):
        self.stop_at_jir = stop_at_jir
        self.jir_count = 0
        self.jirs = {}
        self.memo: Dict[int, Sequence[str]] = {}

    def add_jir(self, jir):
        jir_id = f'm{self.jir_count}'
        self.jir_count += 1
        self.jirs[jir_id] = jir
        return jir_id

    def _add_jir(self, node):
        jir_id = self.add_jir(node._jir)
        if isinstance(node, ir.MatrixIR):
            jref = f'(JavaMatrix {jir_id})'
        elif isinstance(node, ir.TableIR):
            jref = f'(JavaTable {jir_id})'
        elif isinstance(node, ir.BlockMatrixIR):
            jref = f'(JavaBlockMatrix {jir_id})'
        else:
            assert isinstance(node, ir.IR)
            jref = f'(JavaIR {jir_id})'

        self.memo[id(node)] = jref

    def __call__(self, root: 'ir.BaseIR') -> str:
        binding_sites = CSEAnalysisPass(self)(root)
        return CSEPrintPass(self)(root, binding_sites)


class CSEAnalysisPass:
    def __init__(self, renderer: CSERenderer):
        self.renderer = renderer
        self.uid_count = 0

    def uid(self) -> str:
        self.uid_count += 1
        return f'__cse_{self.uid_count}'

    # At top of main loop, we are considering the node 'node' and its
    # 'child_idx'th child, or if 'child_idx' = 'len(node.children)', we are
    # about to do post-processing on 'node' before moving back up to its parent.
    #
    # 'stack' is a stack of 'StackFrame's, one for each node on the path from
    # 'root' to 'node. See 'StackFrame' for descriptions of the maintained state.
    #
    # Returns a Dict summarizing of all lets to be inserted. For each 'node'
    # which will have lets inserted immediately above it, maps 'id(node)' to a
    # 'BindingSite' recording the depth of 'node' and a Dict 'lifted_lets',
    # where for each descendant 'x' which will be bound above 'node',
    # 'lifted_lets' maps 'id(x)' to the unique id 'x' will be bound to.
    def __call__(self, root: 'ir.BaseIR') -> Dict[int, BindingSite]:
        root_frame = self.StackFrame(0, 0, False,
                                     ({var: 0 for var in root.free_vars}, {}, {}),
                                     root)
        stack = [root_frame]
        binding_sites = {}

        while True:
            frame = stack[-1]
            node = frame.node
            frame.child_idx += 1
            child_idx = frame.child_idx

            if child_idx >= len(node.children):
                # mark node as visited at potential let insertion site
                if not node.is_effectful():
                    bind_depth = frame.bind_depth()
                    if bind_depth < frame.min_value_binding_depth:
                        if frame.scan_scope:
                            stack[bind_depth].scan_visited.add(id(node))
                        else:
                            stack[bind_depth].agg_visited.add(id(node))
                    else:
                        stack[bind_depth].visited.add(id(node))

                stack.pop()

                # if any lets being inserted here, add node to registry of
                # binding sites
                if frame.has_lifted_lets():
                    binding_sites[id(node)] = frame.make_binding_site(len(stack))

                if not stack:
                    break
                continue

            child = node.children[child_idx]

            if self.renderer.stop_at_jir and hasattr(child, '_jir'):
                self.renderer._add_jir(child)
                continue

            child_frame = frame.make_child_frame(len(stack))

            if isinstance(child, ir.IR):
                bind_depth = child_frame.bind_depth()
                lets = None
                if bind_depth < len(stack):
                    bind_frame = stack[bind_depth]
                    if bind_depth >= child_frame.min_value_binding_depth:
                        if id(child) in bind_frame.visited:
                            lets = bind_frame.lifted_lets
                    elif child_frame.scan_scope:
                        if id(child) in bind_frame.scan_visited:
                            lets = bind_frame.scan_lifted_lets
                    else:
                        if id(child) in bind_frame.agg_visited:
                            lets = bind_frame.agg_lifted_lets

                # 'lets' is either assigned before one of the 'br/has
                if lets is not None:
                    # we've seen 'child' before, should not traverse (or we will
                    # find too many lifts)
                    if id(child) not in lets:
                        # second time we've seen 'child', lift to a let
                        uid = self.uid()
                        lets[id(child)] = uid
                    continue

            # avoid lifting refs
            if not isinstance(child, ir.Ref):
                # first time visiting 'child'
                stack.append(child_frame)
            continue

        return binding_sites

    class StackFrame:
        __slots__ = ['min_binding_depth', 'min_value_binding_depth', 'scan_scope',
                     'context', 'node', 'visited', 'agg_visited',
                     'scan_visited', 'lifted_lets', 'agg_lifted_lets',
                     'scan_lifted_lets', 'child_idx']

        def __init__(self, min_binding_depth: int, min_value_binding_depth: int,
                     scan_scope: bool, context: Context, x: 'ir.BaseIR'):
            # Immutable:

            # The node corresponding to this stack frame.
            self.node = x
            # If 'node' is to be bound in a let, the let may not rise higher
            # than this (its depth must be >= 'min_binding_depth').
            self.min_binding_depth = min_binding_depth
            # The binding context of 'node'. Maps variables bound above to the
            # depth at which they were bound (more precisely, if
            # 'context[var] == depth', then 'stack[depth-1].node' binds 'var' in
            # the subtree rooted at 'stack[depth].node').
            self.context = context
            # If 'node' is to be bound in a let above this (at a smaller depth)
            # then it must be in an AggLet, with 'scan_scope' determining
            # whether it is bound in the agg scope or scan scope.
            self.min_value_binding_depth = min_value_binding_depth
            self.scan_scope = scan_scope

            # Mutable:

            # Sets of visited descendants. For each descendant 'x' of 'node',
            # 'id(x)' is added to 'visited'. This allows us to recognize when
            # we see a node for a second time.
            # A single descendant may need to be bound here in three separate
            # scopes, so we must track each scope separately.
            self.visited: Set[int] = set()
            self.agg_visited: Set[int] = set()
            self.scan_visited: Set[int] = set()
            # For each descendant 'x' of 'node', if 'x' is to be bound in a Let
            # (resp. an AggLet) immediately above 'node', then 'lifted_lets'
            # (resp. (agg/scan)_lifted_lets) contains 'id(x)', along with the
            # unique id to bind 'x' to.
            self.lifted_lets: Dict[int, str] = {}
            self.agg_lifted_lets: Dict[int, str] = {}
            self.scan_lifted_lets: Dict[int, str] = {}
            # The child currently being visited (satisfies the invariant
            # 'stack[i].node.children[stack[i].child_idx] is stack[i+1].node').
            # Starts at -1 because it is incremented at the top of the main loop.
            self.child_idx = -1

        def has_lifted_lets(self) -> bool:
            return bool(self.lifted_lets or self.agg_lifted_lets or self.scan_lifted_lets)

        def make_binding_site(self, depth):
            return BindingSite(
                lifted_lets=self.lifted_lets,
                agg_lifted_lets=self.agg_lifted_lets,
                scan_lifted_lets=self.scan_lifted_lets,
                depth=depth)

        # compute depth at which we might bind this node
        def bind_depth(self) -> int:
            bind_depth = self.min_binding_depth
            if len(self.node.free_vars) > 0:
                bind_depth = max(bind_depth, max(self.context[0][var] for var in self.node.free_vars))
            if len(self.node.free_agg_vars) > 0:
                bind_depth = max(bind_depth, max(self.context[1][var] for var in self.node.free_agg_vars))
            if len(self.node.free_scan_vars) > 0:
                bind_depth = max(bind_depth, max(self.context[2][var] for var in self.node.free_scan_vars))
            return bind_depth

        def make_child_frame(self, depth: int):
            x = self.node
            i = self.child_idx
            child = x.children[i]
            child_min_binding_depth = self.min_binding_depth
            child_min_value_binding_depth = self.min_value_binding_depth
            child_scan_scope = self.scan_scope
            if x.new_block(i):
                child_min_binding_depth = depth
                child_min_value_binding_depth = depth
            elif x.uses_agg_context(i):
                child_min_value_binding_depth = depth
                child_scan_scope = False
            elif x.uses_scan_context(i):
                child_min_value_binding_depth = depth
                child_scan_scope = True

            child_context = x.child_context(i, self.context, depth)

            return CSEAnalysisPass.StackFrame(child_min_binding_depth, child_min_value_binding_depth, child_scan_scope, child_context, child)


class CSEPrintPass:
    def __init__(self, renderer: CSERenderer):
        self.renderer = renderer

    # At top of main loop, we are considering the 'Renderable' 'node' and its
    # 'child_idx'th child, or if 'child_idx' = 'len(node.children)', we are
    # about to do post-processing on 'node' before moving back up to its parent.
    #
    # 'stack' is a stack of 'StackFrame's, one for each node on the path from
    # 'root' to 'node'. See 'StackFrame' for descriptions of the maintained state.
    #
    # 'bindings_stack' is a stack of 'BindingsStackFrame's, one for each
    # potential binding site on the path from 'root' to 'node'.

    def __call__(self, root: 'ir.BaseIR', binding_sites: Dict[int, BindingSite]):
        root_builder = []
        bindings_stack: Dict[int, CSEPrintPass.BindingsStackFrame] = {}
        memo = self.renderer.memo

        if id(root) in memo:
            return ''.join(memo[id(root)])
        root_ctx = ({var: 0 for var in root.free_vars}, {}, {})
        stack = [self.StackFrame.make(root, self.renderer, binding_sites,
                                      bindings_stack, 0, 0, False, root_ctx, 0)]
        stack[0].set_builder(root_builder, self.renderer)

        while True:
            frame = stack[-1]
            node = frame.node
            frame.child_idx += 1
            child_idx = frame.child_idx

            if child_idx >= len(frame.children):
                if frame.lift_to_frame is not None:
                    assert(not frame.insert_lets)
                    if id(node) in memo:
                        frame.builder.append(memo[id(node)])
                    else:
                        frame.builder.append(node.render_tail(self.renderer))
                    frame.builder.append(' ')
                    # let_bodies is built post-order, which guarantees earlier
                    # lets can't refer to later lets
                    frame.lift_to_frame.let_bodies.append(frame.builder)
                    stack.pop()
                    continue
                else:
                    frame.builder.append(node.render_tail(self.renderer))
                    stack.pop()
                    if frame.insert_lets:
                        if not stack:
                            out_builder = root_builder
                        else:
                            out_builder = stack[-1].builder
                        frame.add_lets(bindings_stack[frame.depth].let_bodies, out_builder)
                        del bindings_stack[frame.depth]
                    if not stack:
                        return ''.join(root_builder)
                    continue

            if child_idx > 0:
                frame.builder.append(' ')
            child = frame.children[child_idx]

            child_frame = frame.make_child_frame(self.renderer, binding_sites, bindings_stack)

            lift_to_frame = None
            if isinstance(child, ir.BaseIR):
                bind_depth = child_frame.bind_depth()
                if bind_depth in bindings_stack:
                    c = bindings_stack[bind_depth]
                    if c.depth >= child_frame.min_value_binding_depth and id(child) in c.lifted_lets:
                        lift_to_frame = c
                        lift_type = 'value'
                    elif child_frame.min_binding_depth <= c.depth < child_frame.min_value_binding_depth:
                        if child_frame.scan_scope and id(child) in c.scan_lifted_lets:
                            lift_to_frame = c
                            lift_type = 'scan'
                        elif not child_frame.scan_scope and id(child) in c.agg_lifted_lets:
                            lift_to_frame = c
                            lift_type = 'agg'

            if lift_to_frame:
                insert_lets = (id(child) in binding_sites
                               and (len(binding_sites[id(child)].lifted_lets) > 0
                                    or len(binding_sites[id(child)].agg_lifted_lets) > 0
                                    or len(binding_sites[id(child)].scan_lifted_lets > 0)))
                assert not insert_lets

                if lift_type == 'value':
                    visited = lift_to_frame.visited
                    name = lift_to_frame.lifted_lets[id(child)]
                elif lift_type == 'agg':
                    visited = lift_to_frame.agg_visited
                    name = lift_to_frame.agg_lifted_lets[id(child)]
                else:
                    visited = lift_to_frame.scan_visited
                    name = lift_to_frame.scan_lifted_lets[id(child)]

                frame.builder.append(f'(Ref {name})')

                if id(child) in visited:
                    continue

                if lift_type == 'value':
                    child_builder = [f'(Let {name} ']
                elif lift_type == 'agg':
                    child_builder = [f'(AggLet {name} False ']
                else:
                    child_builder = [f'(AggLet {name} True ']

                if id(child) in memo:
                    child_builder.append(memo[id(child)])
                    child_builder.append(' ')
                    lift_to_frame.let_bodies.append(child_builder)
                    continue

                visited[id(child)] = child

                child_frame.set_builder(child_builder, self.renderer)
                child_frame.lift_to_frame = lift_to_frame
                stack.append(child_frame)
                continue

            if id(child) in memo:
                frame.builder.append(memo[id(child)])
                continue

            child_frame = frame.make_child_frame(self.renderer, binding_sites, bindings_stack)
            child_frame.set_builder(frame.builder, self.renderer)
            stack.append(child_frame)
            continue

    BindingsStackFrame = namedtuple(
        'BindingsStackFrame',
        'depth lifted_lets agg_lifted_lets scan_lifted_lets visited agg_visited'
        ' scan_visited let_bodies')

    class StackFrame:
        __slots__ = ['node', 'children', 'min_binding_depth', 'context',
                     'min_value_binding_depth', 'scan_scope', 'depth',
                     'lift_to_frame', 'insert_lets', 'builder', 'child_idx']

        def __init__(self,
                     node: Renderable,
                     children: Sequence[Renderable],
                     min_binding_depth: int,
                     min_value_binding_depth: int,
                     scan_scope: bool,
                     context: Context,
                     depth: int,
                     insert_lets: bool,
                     lift_to_frame: 'Optional[CSEPrintPass.BindingsStackFrame]' = None):
            # Immutable

            # The 'Renderable' node corresponding to this stack frame.
            self.node: Renderable = node
            # The list of 'Renderable' children.
            self.children: Sequence[Renderable] = children
            # If 'node' is to be bound in a let, the let may not rise higher
            # than this (its depth must be >= 'min_binding_depth').
            self.min_binding_depth: int = min_binding_depth
            # If 'node' is to be bound higher than this, it must be bound by an
            # AggLet, with 'scan_scope' determining whether it is bound in
            # the agg scope or scan scope.
            self.min_value_binding_depth: int = min_value_binding_depth
            self.scan_scope: bool = scan_scope
            # The binding context of 'node'. Maps variables bound above to the
            # depth at which they were bound (more precisely, if
            # 'context[var] == depth', then 'stack[depth-1].node' binds 'var' in
            # the subtree rooted at 'stack[depth].node').
            self.context = context
            # The depth of 'node' in the original tree, i.e. the number of
            # BaseIR above this in the stack, not counting other 'Renderable's.
            self.depth: int = depth
            # The outermost frame above this in which 'node' was marked to be
            # lifted in the analysis pass, if any, otherwise None.
            self.lift_to_frame: Optional[CSEPrintPass.BindingsStackFrame] = lift_to_frame
            # True if any lets need to be inserted above 'node'. No node has
            # both 'lift_to_frame' not None and 'insert_lets' True.
            self.insert_lets: bool = insert_lets

            # Mutable

            # The index of the 'Renderable' child currently being visited.
            # Starts at -1 because it is incremented at the top of the main loop.
            self.child_idx = -1
            # The array of strings building 'node's IR.
            # * If 'insert_lets', all lets will be added to the parent's
            #   'builder' before appending this 'builder'.
            # * If 'lift_to_frame', 'builder' will be added to 'lift_to_frame's
            #   list of lifted lets, while only "(Ref ...)" will be added to
            #   the parent's 'builder'.
            # * If neither, then it is safe for 'builder' to be an alias of the
            #   parent's 'builder', to save copying.
            self.builder: Optional[MutableSequence[str]] = None

        # compute depth at which we might bind this node
        def bind_depth(self) -> int:
            bind_depth = self.min_binding_depth
            if len(self.node.free_vars) > 0:
                bind_depth = max(bind_depth, max(self.context[0][var] for var in self.node.free_vars))
            if len(self.node.free_agg_vars) > 0:
                bind_depth = max(bind_depth, max(self.context[1][var] for var in self.node.free_agg_vars))
            if len(self.node.free_scan_vars) > 0:
                bind_depth = max(bind_depth, max(self.context[2][var] for var in self.node.free_scan_vars))
            return bind_depth

        def add_lets(self, let_bodies: Sequence[str], out_builder: MutableSequence[str]):
            for let_body in let_bodies:
                out_builder.extend(let_body)
            out_builder.extend(self.builder)
            num_lets = len(let_bodies)
            for _ in range(num_lets):
                out_builder.append(')')

        def make_child_frame(self,
                             renderer: 'CSERenderer',
                             binding_sites: Dict[int, BindingSite],
                             bindings_stack: 'Dict[int, CSEPrintPass.BindingsStackFrame]'):
            child_min_binding_depth = self.min_binding_depth
            child_min_value_binding_depth = self.min_value_binding_depth
            child_scan_scope = self.scan_scope

            if isinstance(self.node, ir.BaseIR):
                if self.node.renderable_new_block(self.child_idx):
                    child_min_binding_depth = self.depth + 1
                    child_min_value_binding_depth = self.depth + 1
                elif self.node.renderable_uses_agg_context(self.child_idx):
                    child_min_value_binding_depth = self.depth + 1
                    child_scan_scope = False
                elif self.node.renderable_uses_scan_context(self.child_idx):
                    child_min_value_binding_depth = self.depth + 1
                    child_scan_scope = True

            child = self.children[self.child_idx]
            if isinstance(child, ir.BaseIR):
                child_depth = self.depth + 1
            else:
                child_depth = self.depth
            if isinstance(self.node, ir.BaseIR):
                child_context = self.node.renderable_child_context(self.child_idx, self.context, child_depth)
            else:
                child_context = self.context
            return self.make(child, renderer, binding_sites, bindings_stack,
                             child_min_binding_depth, child_min_value_binding_depth,
                             child_scan_scope, child_context, child_depth)

        @staticmethod
        def make(node: Renderable,
                 renderer: 'CSERenderer',
                 binding_sites: Dict[int, BindingSite],
                 bindings_stack: 'Dict[int, CSEPrintPass.BindingsStackFrame]',
                 min_binding_depth: int,
                 min_value_binding_depth: int,
                 scan_scope: bool,
                 context: Context,
                 depth: int):
            insert_lets = (id(node) in binding_sites
                           and depth == binding_sites[id(node)].depth
                           and (len(binding_sites[id(node)].lifted_lets) > 0
                                or len(binding_sites[id(node)].agg_lifted_lets) > 0
                                or len(binding_sites[id(node)].scan_lifted_lets) > 0))
            state = CSEPrintPass.StackFrame(node,
                                            node.render_children(renderer),
                                            min_binding_depth,
                                            min_value_binding_depth,
                                            scan_scope, context, depth, insert_lets)
            if insert_lets:
                bind_site = binding_sites[id(node)]
                assert(bind_site.depth == depth)
                bindings_stack[depth] = CSEPrintPass.StackFrame.make_bindings_stack_frame(bind_site)
            return state

        def set_builder(self, builder, renderer: 'CSERenderer'):
            self.builder = builder
            if self.insert_lets:
                self.builder = []
            head = self.node.render_head(renderer)
            if head != '':
                self.builder.append(head)

        @staticmethod
        def make_bindings_stack_frame(site: BindingSite):
            return CSEPrintPass.BindingsStackFrame(
                depth=site.depth,
                lifted_lets=site.lifted_lets,
                agg_lifted_lets=site.agg_lifted_lets,
                scan_lifted_lets=site.scan_lifted_lets,
                visited={},
                agg_visited={},
                scan_visited={},
                let_bodies=[])



from hail.expr.expressions import *
from hail.expr.expressions.expression_typecheck import *
from hail.expr.types import *
from hail.typecheck import *
from hail.expr.functions import _func


@typecheck(gt_counts=expr_array(expr_int32))
def haplotype_freq_em(gt_counts) -> ArrayExpression:
    """
    Computes estimated haplotype counts based on genotypes for a pair of bi-allelic variants.
    Implements the Excoffier & Slatkin EM (Exccoffier & Slatkin, Mol. Biol. Evol. 1995)

    The unphased input genotype counts for the variant pairs has to be provided in the following order:
    [AABB, AABb, AAbb, AaBB, AaBb, Aabb, aaBB, aaBb, aabb]

    The estimated haplotype counts are returned in an array in the following order:
    [AB, Ab, aB, ab]

    Where _A_ and _a_ are the reference and non-reference alleles for the first variant, resp.
    And _B_ and _b_ are the reference and non-reference alleles for the second variant, resp.

    Parameters
    ----------
    gt_counts : :class:`.ArrayExpression`

    Returns
    -------
    :class:`.ArrayExpression`
    """
    return _func("haplotype_freq_em", tarray(tfloat64), gt_counts)


from IPython.display import display
from ipywidgets import widgets

import hail as hl

from hail.expr.types import summary_type

__all__ = [
    'interact',
]

def interact(obj):
    tab = widgets.Tab()
    base_style = widgets.ButtonStyle()
    selected_style = widgets.ButtonStyle(button_color='#DDFFDD', font_weight='bold')

    if isinstance(obj, hl.Table):
        glob = widgets.Button(description='globals',
                              layout=widgets.Layout(width='150px', height='30px'))
        rows = widgets.Button(description='rows',
                              layout=widgets.Layout(width='150px', height='200px'))
        rows.style = selected_style

        globals_frames = []
        globals_frames.append(widgets.HTML(
            f'<p><big>Global fields, with one value in the dataset.</big></p>\n'
            f'<p>Commonly used methods:</p>\n'
            f'<ul>'
            f'<li>{html_link("annotate_globals", "https://hail.is/docs/0.2/hail.Table.html#hail.Table.annotate_globals")}: '
            f'add new global fields.</li>'
            f'</ul>'
        ))
        append_struct_frames(obj.globals.dtype, globals_frames)

        row_frames = []
        row_frames.append(widgets.HTML(
            f'<p><big>Row fields, with one record per row of the table.</big></p>\n'
            f'<p>Commonly used methods:</p>\n'
            f'<ul>'
            f'<li>{html_link("annotate", "https://hail.is/docs/0.2/hail.Table.html#hail.Table.annotate")}: '
            f'add new fields.</li>'
            f'<li>{html_link("filter", "https://hail.is/docs/0.2/hail.Table.html#hail.Table.filter")}: '
            f'filter rows of the table.</li>'
            f'<li>{html_link("aggregate", "https://hail.is/docs/0.2/hail.Table.html#hail.Table.aggregate")}: '
            f'aggregate over rows to produce a single value.</li>'
            f'</ul>'
        ))
        if len(obj.key) > 0:
            row_frames.append(widgets.HTML(f'<p><big>Key: {list(obj.key)}<big><p>'))
        append_struct_frames(obj.row.dtype, row_frames)

        tab.children = [widgets.VBox(frames) for frames in [globals_frames, row_frames]]
        tab.set_title(0, 'globals')
        tab.set_title(1, 'row')
        tab.selected_index = 1

        box = widgets.VBox([glob, rows])
        buttons = [glob, rows]
    else:
        assert isinstance(obj, hl.MatrixTable)
        glob = widgets.Button(description='globals',
                              layout=widgets.Layout(width='65px', height='30px'))
        cols = widgets.Button(description='cols',
                              layout=widgets.Layout(width='200px', height='30px'))
        rows = widgets.Button(description='rows',
                              layout=widgets.Layout(width='65px', height='200px'))
        entries = widgets.Button(description='entries',
                                 layout=widgets.Layout(width='200px', height='200px'))
        entries.style = selected_style

        globals_frames = []
        globals_frames.append(widgets.HTML(
            f'<p><big>Global fields, with one value in the dataset.</big></p>\n'
            f'<p>Commonly used methods:</p>\n'
            f'<ul>'
            f'<li>{html_link("annotate_globals()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.annotate_globals")}: '
            f'add new global fields.</li>'
            f'</ul>'
        ))
        append_struct_frames(obj.globals.dtype, globals_frames)

        row_frames = []
        row_frames.append(widgets.HTML(
            f'<p><big>Row fields, with one record per row in the dataset.</big></p>\n'
            f'<p>Commonly used methods:</p>\n'
            f'<ul>'
            f'<li>{html_link("annotate_rows()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.annotate_rows")}: '
            f'add new row fields. This method supports {html_link("aggregation", "https://hail.is/docs/0.2/aggregators.html")}, '
            f'aggregating over entries to compute one result per row, e.g. computing the mean depth per variant.</li>'
            f'<li>{html_link("filter_rows()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.filter_rows")}: '
            f'filter rows in the matrix table.</li>'
            f'<li>{html_link("aggregate_rows()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.aggregate_rows")}: '
            f'aggregate over rows (not including entries or columns) to produce a single value, e.g. counting the number of loss-of-function variants.</li>'
            f'<li>{html_link("rows()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.rows")}: '
            f'return the rows as a Hail {html_link("Table", "https://hail.is/docs/0.2/hail.Table.html")}.</li>'
            f'</ul>'
        ))
        if len(obj.row_key) > 0:
            row_frames.append(widgets.HTML(f'<p><big>Row key: {list(obj.row_key)}<big><p>'))
        append_struct_frames(obj.row.dtype, row_frames)

        col_frames = []
        col_frames.append(widgets.HTML(
            f'<p><big>Column fields, with one record per column in the dataset.</big></p>\n'
            f'<p>Commonly used methods:</p>\n'
            f'<ul>'
            f'<li>{html_link("annotate_cols()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.annotate_cols")}: '
            f'add new column fields. This method supports {html_link("aggregation", "https://hail.is/docs/0.2/aggregators.html")}, '
            f'aggregating over entries to compute one result per column, e.g. computing the mean depth per sample.</li>'
            f'<li>{html_link("filter_cols()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.filter_cols")}: '
            f'filter columns in the matrix table.</li>'
            f'<li>{html_link("aggregate_cols()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.aggregate_cols")}: '
            f'aggregate over columns (not including entries or rows) to produce a single value, e.g. counting the number of samples with case status.</li>'
            f'<li>{html_link("cols()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.cols")}: '
            f'return the columns as a Hail {html_link("Table", "https://hail.is/docs/0.2/hail.Table.html")}.'
            f'</li>'
            f'</ul>'
        ))
        if len(obj.col_key) > 0:
            col_frames.append(widgets.HTML(f'<p><big>Column key: {list(obj.col_key)}<big><p>'))
        append_struct_frames(obj.col.dtype, col_frames)

        entry_frames = []
        entry_frames.append(widgets.HTML(
            f'<p><big>Entry fields, with one record per (row, column) pair in the dataset.</big></p>\n'
            f'<p>Commonly used methods:</p>\n'
            f'<ul>'
            f'<li>{html_link("annotate_entries()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.annotate_entries")}: '
            f'add new entry fields.</li>'
            f'<li>{html_link("filter_entries()", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.filter_entries")}: '
            f'filter entries in the matrix table, removing them from downstream operations, like aggregations.</li>'
            f'<li>{html_link("aggregate_entries", "https://hail.is/docs/0.2/hail.MatrixTable.html#hail.MatrixTable.aggregate_entries")}: '
            f'aggregate over entries to produce a single value, e.g. computing mean depth across an entire dataset.</li>'
            f'</ul>'
        ))
        append_struct_frames(obj.entry.dtype, entry_frames)

        tab.children = [widgets.VBox(frames) for frames in [globals_frames, row_frames, col_frames, entry_frames]]
        tab.set_title(0, 'globals')
        tab.set_title(1, 'row')
        tab.set_title(2, 'col')
        tab.set_title(3, 'entry')
        tab.selected_index = 3

        box = widgets.VBox([widgets.HBox([glob, cols]), widgets.HBox([rows, entries])])
        buttons = [glob, rows, cols, entries]

    selection_handler = widgets.IntText(tab.selected_index)
    button_idx = dict(zip(buttons, range(len(buttons))))

    def handle_selection(x):
        if x['name'] == 'value' and x['type'] == 'change':
            buttons[x['old']].style = base_style
            selection = x['new']
            buttons[selection].style = selected_style
            tab.selected_index = selection

    selection_handler.observe(handle_selection)
    widgets.jslink((tab, 'selected_index'), (selection_handler, 'value'))

    def button_action(b):
        selection_handler.value = button_idx[b]

    for button in button_idx:
        button.on_click(button_action)

    display(box, tab)

def html_code(text):
    return f'<pre>{text}</pre>'


def get_type_html(t):
    if isinstance(t, hl.tdict):
        return f'<p>A dictionary mapping keys to values.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" DictExpression", "https://hail.is/docs/0.2/hail.expr.DictExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" CollectionExpression", "https://hail.is/docs/0.2/hail.expr.CollectionExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'</ul>\n' \
               f'<p>Access elements using square brackets: {html_code("x[k]")}</p>'
    elif isinstance(t, hl.tset):
        return f'<p>A set of unique values.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" SetExpression", "https://hail.is/docs/0.2/hail.expr.SetExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" CollectionExpression", "https://hail.is/docs/0.2/hail.expr.CollectionExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'</ul>\n'
    elif isinstance(t, hl.tarray):
        if hl.expr.types.is_numeric(t.element_type):
            return f'<p>A variable-length array of homogenous numeric values.</p>\n' \
                   f'Documentation:\n<ul>' \
                   f'<li>class: {html_link(" ArrayNumericExpression", "https://hail.is/docs/0.2/hail.expr.ArrayNumericExpression.html")}</li>' \
                   f'<li>inherited class: {html_link(" ArrayExpression", "https://hail.is/docs/0.2/hail.expr.ArrayExpression.html")}</li>' \
                   f'<li>inherited class: {html_link(" CollectionExpression", "https://hail.is/docs/0.2/hail.expr.CollectionExpression.html")}</li>' \
                   f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
                   f'</ul>\n' \
                   f'<p>Access elements using square brackets: {html_code("x[i]")} or slice using Python syntax: {html_code("x[:end], x[start:], x[start:end]")}</p>'
        else:
            return f'<p>A variable-length array of homogenous values.</p>\n' \
                   f'Documentation:\n<ul>' \
                   f'<li>class: {html_link(" ArrayExpression", "https://hail.is/docs/0.2/hail.expr.ArrayExpression.html")}</li>' \
                   f'<li>inherited class: {html_link(" CollectionExpression", "https://hail.is/docs/0.2/hail.expr.CollectionExpression.html")}</li>' \
                   f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
                   f'</ul>\n' \
                   f'<p>Access elements using square brackets: {html_code("x[i]")} or slice using Python syntax: {html_code("x[:end], x[start:], x[start:end]")}</p>'
    elif isinstance(t, hl.tstruct):
        bracket_str = html_code("x[\"foo\"]")
        return f'<p>A structure of named heterogeneous values.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" StructExpression", "https://hail.is/docs/0.2/hail.expr.StructExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'</ul>\n' \
               f'<p>Access an element by name with dots or with square brackets: {html_code(f"x.foo, {bracket_str}")}</p>'
    elif isinstance(t, hl.ttuple):
        return f'<p>A 0-indexed tuple of heterogeneous values.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" TupleExpression", "https://hail.is/docs/0.2/hail.expr.TupleExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'</ul>\n' \
               f'<p>Access an element using square brackets. For instance, get the first element: {html_code("x[0]")}</p>'
    elif isinstance(t, hl.tinterval):
        return f'<p>An object representing an interval.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" IntervalExpression", "https://hail.is/docs/0.2/hail.expr.IntervalExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'</ul>'
    elif isinstance(t, hl.tlocus):
        return f'<p>An object representing a genomic locus (chromomsome and position).</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" LocusExpression", "https://hail.is/docs/0.2/hail.expr.LocusExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'<li>{html_link("Genetics functions", "https://hail.is/docs/0.2/functions/genetics.html")}</li>' \
               f'</ul>'
    elif t == hl.tint32:
        return f'<p>A 32-bit integer.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" Int32Expression", "https://hail.is/docs/0.2/hail.expr.Int32Expression.html")}</li>' \
               f'<li>inherited class: {html_link(" NumericExpression", "https://hail.is/docs/0.2/hail.expr.NumericExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'<li>{html_link("Numeric functions", "https://hail.is/docs/0.2/functions/numeric.html")}</li>' \
               f'<li>{html_link("Statistical functions", "https://hail.is/docs/0.2/functions/stats.html")}</li>' \
               f'</ul>'
    elif t == hl.tint64:
        return f'<p>A 64-bit integer.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" Int64Expression", "https://hail.is/docs/0.2/hail.expr.Int64Expression.html")}</li>' \
               f'<li>inherited class: {html_link(" NumericExpression", "https://hail.is/docs/0.2/hail.expr.NumericExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'<li>{html_link("Numeric functions", "https://hail.is/docs/0.2/functions/numeric.html")}</li>' \
               f'<li>{html_link("Statistical functions", "https://hail.is/docs/0.2/functions/stats.html")}</li>' \
               f'</ul>'
    elif t == hl.tfloat32:
        return f'<p>A 32-bit floating point number.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" Float32Expression", "https://hail.is/docs/0.2/hail.expr.Float32Expression.html")}</li>' \
               f'<li>inherited class: {html_link(" NumericExpression", "https://hail.is/docs/0.2/hail.expr.NumericExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'<li>{html_link("Numeric functions", "https://hail.is/docs/0.2/functions/numeric.html")}</li>' \
               f'<li>{html_link("Statistical functions", "https://hail.is/docs/0.2/functions/stats.html")}</li>' \
               f'</ul>'
    elif t == hl.tfloat64:
        return f'<p>A 64-bit floating point number.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" Float64Expression", "https://hail.is/docs/0.2/hail.expr.Float64Expression.html")}</li>' \
               f'<li>inherited class: {html_link(" NumericExpression", "https://hail.is/docs/0.2/hail.expr.NumericExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'<li>{html_link("Numeric functions", "https://hail.is/docs/0.2/functions/numeric.html")}</li>' \
               f'<li>{html_link("Statistical functions", "https://hail.is/docs/0.2/functions/stats.html")}</li>' \
               f'</ul>'
    elif t == hl.tbool:
        return f'<p>A 64-bit floating point number.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" BooleanExpression", "https://hail.is/docs/0.2/hail.expr.BooleanExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" NumericExpression", "https://hail.is/docs/0.2/hail.expr.NumericExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'<li>{html_link("Numeric functions", "https://hail.is/docs/0.2/functions/numeric.html")}</li>' \
               f'<li>{html_link("Statistical functions", "https://hail.is/docs/0.2/functions/stats.html")}</li>' \
               f'</ul>'

    elif t == hl.tstr:
        return f'<p>A text string.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" StringExpression", "https://hail.is/docs/0.2/hail.expr.StringExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'</ul>'
    elif t == hl.tcall:
        return f'<p>An object representing a genotype call.</p>\n' \
               f'Documentation:\n<ul>' \
               f'<li>class: {html_link(" CallExpression", "https://hail.is/docs/0.2/hail.expr.CallExpression.html")}</li>' \
               f'<li>inherited class: {html_link(" Expression", "https://hail.is/docs/0.2/hail.expr.Expression.html")}</li>' \
               f'</ul>'


def html_link(text, dest):
    return f'<a target="_blank" href="{dest}">{text}</a>'


def format_html(s):
    return '<p style="font-family:courier;white-space:pre;line-height: 115%;">{}</p>'.format(
        str(s).replace('<', '&lt').replace('>', '&gt').replace('\n', '</br>'))


def append_struct_frames(t, frames):
    if len(t) == 0:
        frames.append(widgets.HTML('<big>No fields.</big>'))
    else:
        frames.append(widgets.HTML('<big>Fields:</big>'))
    acc = widgets.Accordion([recursive_build(x) for x in t.values()])
    for i, (name, fd) in enumerate(t.items()):
        acc.set_title(i, f'{repr(name)} ({summary_type(fd)})')
    acc.selected_index = None
    frames.append(acc)


def recursive_build(t):
    frames = []

    frames.append(widgets.HTML(get_type_html(t)))

    if isinstance(t, hl.tstruct):
        append_struct_frames(t, frames)
    elif isinstance(t, hl.ttuple):
        if len(t) == 0:
            frames.append(widgets.HTML('<big>No fields.</big>'))
        else:
            frames.append(widgets.HTML('<big>Fields:</big>'))
        acc = widgets.Accordion([recursive_build(x) for x in t.types])
        for i, fd in enumerate(t.types):
            acc.set_title(i, f'[{i}] ({summary_type(fd)})')
        acc.selected_index = None
        frames.append(acc)
    elif isinstance(t, (hl.tarray, hl.tset)):
        acc = widgets.Accordion([recursive_build(t.element_type)])
        acc.set_title(0, f'<element> ({summary_type(t.element_type)})')
        acc.selected_index = None
        frames.append(acc)
    elif isinstance(t, hl.tdict):
        acc = widgets.Accordion([recursive_build(t.key_type), recursive_build(t.value_type)])
        acc.set_title(0, f'<key> ({summary_type(t.key_type)})')
        acc.set_title(1, f'<value> ({summary_type(t.element_type)})')
        acc.selected_index = None
        frames.append(acc)
    elif isinstance(t, (hl.tinterval)):
        acc = widgets.Accordion([recursive_build(t.point_type)])
        acc.set_title(0, f'<point> ({summary_type(t.point_type)})')
        acc.selected_index = None
        frames.append(acc)

    return widgets.VBox(frames)



from typing import *

import numpy as np

from hail.expr import expressions
from hail.expr.types import *
from hail.expr.types import from_numpy
from hail.ir import *
from hail.typecheck import linked_list
from hail.utils.java import *
from hail.utils.linkedlist import LinkedList
from .indices import *

from hail.expr.types import summary_type

class Summary(object):
    def __init__(self, type, count, summ_fields, nested, header=None):
        self.count = count
        self.summ_fields = summ_fields
        self.nested = nested
        self.type = type
        self.header = header

    @staticmethod
    def pct(x):
        return f'{x*100:.2f}%'

    @staticmethod
    def format(x):
        if isinstance(x, float):
            return f'{x:.2f}'
        else:
            return str(x)

    def __str__(self):
        return self._ascii_string(depth=0, prefix=None)

    def __repr__(self):
        return self.__str__()

    def _repr_html_(self):
        return self._html_string(prefix=None)

    def _ascii_string(self, depth, prefix):
        spacing = '  ' * depth

        summary = ''
        if self.header:
            summary += f'\n{spacing}{self.header}'

        if prefix is not None:
            summary += f'\n\n{spacing}- {prefix} ({summary_type(self.type)}):'

        if len(self.summ_fields) > 0:
            max_n_len = max(len(n) for n in self.summ_fields.keys())
            for name, v in self.summ_fields.items():
                summary += f'\n{spacing}  {name.rjust(max_n_len)}: {self.format(v)}'
        for name, field in self.nested.items():
            if prefix is not None:
                name = f'{prefix}{name}'
            summary += field._ascii_string(depth + 1, prefix=name)

        return summary

    def _html_string(self, prefix):
        import html
        summary = ''
        if self.header:
            summary += f'<p>{self.header}</p>'
        if prefix is not None:
            summary += f'<b>{html.escape(prefix)}</b> (<i>{html.escape(summary_type(self.type))}</i>):'

        summary += '<ul>'
        if len(self.summ_fields) > 0:
            summary += '<table><tbody>'
            for name, v in self.summ_fields.items():
                summary += f'<tr><td>{html.escape(name)}</td><td>{html.escape(self.format(v))}</td></tr>'
            summary += '</tbody></table>'
        for name, field in self.nested.items():
            if prefix is not None:
                name = f'{prefix}{name}'
            summary += '<li>' + field._html_string(prefix=name) + '</li>'
        summary += '</ul>'

        return summary


class NamedSummary(object):
    def __init__(self, summary, name, header):
        self.header = header
        self.name = name
        self.summary = summary

    def __str__(self):
        s = self.summary._ascii_string(depth=0, prefix=self.name)
        if self.header:
            s = f'{self.header}\n======' + s
        return s

    def __repr__(self):
        return self.__str__()

    def _repr_html_(self):
        import html
        s = self.summary._html_string(prefix=self.name)
        if self.header:
            s = f'<h3>{html.escape(self.header)}</h3>' + s
        return s


class ExpressionException(Exception):
    def __init__(self, msg=''):
        self.msg = msg
        super(ExpressionException, self).__init__(msg)


class ExpressionWarning(Warning):
    def __init__(self, msg=''):
        self.msg = msg
        super(ExpressionWarning, self).__init__(msg)


def impute_type(x):
    from hail.genetics import Locus, Call
    from hail.utils import Interval, Struct
    
    if isinstance(x, Expression):
        return x.dtype
    elif isinstance(x, bool):
        return tbool
    elif isinstance(x, int):
        if hl.tint32.min_value <= x <= hl.tint32.max_value:
            return tint32
        elif hl.tint64.min_value <= x <= hl.tint64.max_value:
            return tint64
        else:
            raise ValueError("Hail has no integer data type large enough to store {}".format(x))
    elif isinstance(x, float):
        return tfloat64
    elif isinstance(x, str):
        return tstr
    elif isinstance(x, Locus):
        return tlocus(x.reference_genome)
    elif isinstance(x, Interval):
        return tinterval(x.point_type)
    elif isinstance(x, Call):
        return tcall
    elif isinstance(x, Struct):
        return tstruct(**{k: impute_type(x[k]) for k in x})
    elif isinstance(x, tuple):
        return ttuple(*(impute_type(element) for element in x))
    elif isinstance(x, list):
        if len(x) == 0:
            raise ExpressionException("Cannot impute type of empty list. Use 'hl.empty_array' to create an empty array.")
        ts = {impute_type(element) for element in x}
        unified_type = unify_types_limited(*ts)
        if not unified_type:
            raise ExpressionException("Hail does not support heterogeneous arrays: "
                                      "found list with elements of types {} ".format(list(ts)))
        return tarray(unified_type)
    elif isinstance(x, set):
        if len(x) == 0:
            raise ExpressionException("Cannot impute type of empty set. Use 'hl.empty_set' to create an empty set.")
        ts = {impute_type(element) for element in x}
        unified_type = unify_types_limited(*ts)
        if not unified_type:
            raise ExpressionException("Hail does not support heterogeneous sets: "
                                      "found set with elements of types {} ".format(list(ts)))
        return tset(unified_type)
    elif isinstance(x, dict):
        if len(x) == 0:
            raise ExpressionException("Cannot impute type of empty dict. Use 'hl.empty_dict' to create an empty dict.")
        kts = {impute_type(element) for element in x.keys()}
        vts = {impute_type(element) for element in x.values()}
        unified_key_type = unify_types_limited(*kts)
        unified_value_type = unify_types_limited(*vts)
        if not unified_key_type:
            raise ExpressionException("Hail does not support heterogeneous dicts: "
                                      "found dict with keys of types {} ".format(list(kts)))
        if not unified_value_type:
            raise ExpressionException("Hail does not support heterogeneous dicts: "
                                      "found dict with values of types {} ".format(list(vts)))
        return tdict(unified_key_type, unified_value_type)
    elif isinstance(x, np.generic):
        return from_numpy(x.dtype)
    elif isinstance(x, np.ndarray):
        element_type = from_numpy(x.dtype)
        return tndarray(element_type, x.ndim)
    elif x is None:
        raise ExpressionException("Hail cannot impute the type of 'None'")
    elif isinstance(x, (hl.expr.builders.CaseBuilder, hl.expr.builders.SwitchBuilder)):
        raise ExpressionException("'switch' and 'case' expressions must end with a call to either"
                                  "'default' or 'or_missing'")
    else:
        raise ExpressionException("Hail cannot automatically impute type of {}: {}".format(type(x), x))


def to_expr(e, dtype=None) -> 'Expression':
    if isinstance(e, Expression):
        if dtype and not dtype == e.dtype:
            raise TypeError("expected expression of type '{}', found expression of type '{}'".format(dtype, e.dtype))
        return e
    return cast_expr(e, dtype)

def cast_expr(e, dtype=None) -> 'Expression':
    if not dtype:
        dtype = impute_type(e)
    x = _to_expr(e, dtype)
    if isinstance(x, Expression):
        return x
    else:
        return hl.literal(x, dtype)


def _to_expr(e, dtype):
    if e is None:
        return None
    elif isinstance(e, Expression):
        if e.dtype != dtype:
            assert is_numeric(dtype), 'expected {}, got {}'.format(dtype, e.dtype)
            if dtype == tfloat64:
                return hl.float64(e)
            elif dtype == tfloat32:
                return hl.float32(e)
            elif dtype == tint64:
                return hl.int64(e)
            else:
                assert dtype == tint32
                return hl.int32(e)
        return e
    elif not is_compound(dtype):
        # these are not container types and cannot contain expressions if we got here
        return e
    elif isinstance(dtype, tstruct):
        new_fields = []
        found_expr = False
        for f, t in dtype.items():
            value = _to_expr(e[f], t)
            found_expr = found_expr or isinstance(value, Expression)
            new_fields.append(value)

        if not found_expr:
            return e
        else:
            exprs = [new_fields[i] if isinstance(new_fields[i], Expression)
                     else hl.literal(new_fields[i], dtype[i])
                     for i in range(len(new_fields))]
            fields = {name: expr for name, expr in zip(dtype.keys(), exprs)}
            from .typed_expressions import StructExpression
            return StructExpression._from_fields(fields)

    elif isinstance(dtype, tarray):
        elements = []
        found_expr = False
        for element in e:
            value = _to_expr(element, dtype.element_type)
            found_expr = found_expr or isinstance(value, Expression)
            elements.append(value)
        if not found_expr:
            return e
        else:
            assert (len(elements) > 0)
            exprs = [element if isinstance(element, Expression)
                     else hl.literal(element, dtype.element_type)
                     for element in elements]
            indices, aggregations = unify_all(*exprs)
        ir = MakeArray([e._ir for e in exprs], None)
        return expressions.construct_expr(ir, dtype, indices, aggregations)
    elif isinstance(dtype, tset):
        elements = []
        found_expr = False
        for element in e:
            value = _to_expr(element, dtype.element_type)
            found_expr = found_expr or isinstance(value, Expression)
            elements.append(value)
        if not found_expr:
            return e
        else:
            assert (len(elements) > 0)
            exprs = [element if isinstance(element, Expression)
                     else hl.literal(element, dtype.element_type)
                     for element in elements]
            indices, aggregations = unify_all(*exprs)
            ir = ToSet(ToStream(MakeArray([e._ir for e in exprs], None)))
            return expressions.construct_expr(ir, dtype, indices, aggregations)
    elif isinstance(dtype, ttuple):
        elements = []
        found_expr = False
        assert len(e) == len(dtype.types)
        for i in range(len(e)):
            value = _to_expr(e[i], dtype.types[i])
            found_expr = found_expr or isinstance(value, Expression)
            elements.append(value)
        if not found_expr:
            return e
        else:
            exprs = [elements[i] if isinstance(elements[i], Expression)
                     else hl.literal(elements[i], dtype.types[i])
                     for i in range(len(elements))]
            indices, aggregations = unify_all(*exprs)
            ir = MakeTuple([expr._ir for expr in exprs])
            return expressions.construct_expr(ir, dtype, indices, aggregations)
    elif isinstance(dtype, tdict):
        keys = []
        values = []
        found_expr = False
        for k, v in e.items():
            k_ = _to_expr(k, dtype.key_type)
            v_ = _to_expr(v, dtype.value_type)
            found_expr = found_expr or isinstance(k_, Expression)
            found_expr = found_expr or isinstance(v_, Expression)
            keys.append(k_)
            values.append(v_)
        if not found_expr:
            return e
        else:
            assert len(keys) > 0
            # Here I use `to_expr` to call `lit` the keys and values separately.
            # I anticipate a common mode is statically-known keys and Expression
            # values.
            key_array = to_expr(keys, tarray(dtype.key_type))
            value_array = to_expr(values, tarray(dtype.value_type))
            return hl.dict(hl.zip(key_array, value_array))
    elif isinstance(dtype, hl.tndarray):
        return hl.nd.array(e)
    else:
        raise NotImplementedError(dtype)


def unify_all(*exprs) -> Tuple[Indices, LinkedList]:
    if len(exprs) == 0:
        return Indices(), LinkedList(Aggregation)
    try:
        new_indices = Indices.unify(*[e._indices for e in exprs])
    except ExpressionException:
        # source mismatch
        from collections import defaultdict
        sources = defaultdict(lambda: [])
        for e in exprs:
            from .expression_utils import get_refs
            for name, inds in get_refs(e, *[e for a in e._aggregations for e in a.exprs]).items():
                sources[inds.source].append(str(name))
        raise ExpressionException("Cannot combine expressions from different source objects."
                                  "\n    Found fields from {n} objects:{fields}".format(
            n=len(sources),
            fields=''.join("\n        {}: {}".format(src, fds) for src, fds in sources.items())
        )) from None
    first, rest = exprs[0], exprs[1:]
    aggregations = first._aggregations
    for e in rest:
        aggregations = aggregations.push(*e._aggregations)
    return new_indices, aggregations


def unify_types_limited(*ts):
    type_set = set(ts)
    if len(type_set) == 1:
        # only one distinct class
        return next(iter(type_set))
    elif all(is_numeric(t) for t in ts):
        # assert there are at least 2 numeric types
        assert len(type_set) > 1
        if tfloat64 in type_set:
            return tfloat64
        elif tfloat32 in type_set:
            return tfloat32
        elif tint64 in type_set:
            return tint64
        else:
            assert type_set == {tint32, tbool}
            return tint32
    else:
        return None


def unify_types(*ts):
    limited_unify = unify_types_limited(*ts)
    if limited_unify is not None:
        return limited_unify
    elif all(isinstance(t, tarray) for t in ts):
        et = unify_types_limited(*(t.element_type for t in ts))
        if et is not None:
            return tarray(et)
        else:
            return None
    else:
        return None

def unify_exprs(*exprs: 'Expression') -> Tuple:
    assert len(exprs) > 0
    types = {e.dtype for e in exprs}

    # all types are the same
    if len(types) == 1:
        return exprs + (True,)

    for t in types:
        c = expressions.coercer_from_dtype(t)
        if all(c.can_coerce(e.dtype) for e in exprs):
            return tuple([c.coerce(e) for e in exprs]) + (True,)

    # cannot coerce all types to the same type
    return exprs + (False,)

class Expression(object):
    """Base class for Hail expressions."""

    __array_ufunc__ = None  # disable NumPy coercions, so Hail coercions take priority

    @typecheck_method(ir=IR, type=nullable(HailType), indices=Indices, aggregations=linked_list(Aggregation))
    def __init__(self,
                 ir: IR,
                 type: HailType,
                 indices: Indices = Indices(),
                 aggregations: LinkedList = LinkedList(Aggregation)):

        self._ir: IR = ir
        self._type = type
        self._indices = indices
        self._aggregations = aggregations
        self._summary = None

    def describe(self, handler=print):
        """Print information about type, index, and dependencies."""
        if self._aggregations:
            agg_indices = set()
            for a in self._aggregations:
                agg_indices = agg_indices.union(a.indices.axes)
            agg_tag = ' (aggregated)'
            agg_str = f'Includes aggregation with index {list(agg_indices)}\n' \
                      f'    (Aggregation index may be promoted based on context)'
        else:
            agg_tag = ''
            agg_str = ''

        bar = '--------------------------------------------------------'
        s = '{bar}\n' \
            'Type:\n' \
            '    {t}\n' \
            '{bar}\n' \
            'Source:\n' \
            '    {src}\n' \
            'Index:\n' \
            '    {inds}{agg_tag}{maybe_bar}{agg}\n' \
            '{bar}'.format(bar=bar,
                           t=self.dtype.pretty(indent=4),
                           src=self._indices.source,
                           inds=list(self._indices.axes),
                           maybe_bar='\n' + bar + '\n' if agg_str else '',
                           agg_tag=agg_tag,
                           agg=agg_str)
        handler(s)

    def __lt__(self, other):
        return self._compare_op("<", other)

    def __le__(self, other):
        return self._compare_op("<=", other)

    def __gt__(self, other):
        return self._compare_op(">", other)

    def __ge__(self, other):
        return self._compare_op(">=", other)

    def __nonzero__(self):
        raise ExpressionException(
            "The truth value of an expression is undefined\n"
            "    Hint: instead of 'if x', use 'hl.cond(x, ...)'\n"
            "    Hint: instead of 'x and y' or 'x or y', use 'x & y' or 'x | y'\n"
            "    Hint: instead of 'not x', use '~x'")

    def __iter__(self):
        raise ExpressionException(f"{repr(self)} object is not iterable")

    def _compare_op(self, op, other):
        other = to_expr(other)
        left, right, success = unify_exprs(self, other)
        if not success:
            raise TypeError(f"Invalid '{op}' comparison, cannot compare expressions "
                            f"of type '{self.dtype}' and '{other.dtype}'")
        res = left._bin_op(op, right, hl.tbool)
        return res

    def _is_scalar(self):
        return self._indices.source is None

    def _promote_scalar(self, typ):
        if typ == tint32:
            return hail.int32(self)
        elif typ == tint64:
            return hail.int64(self)
        elif typ == tfloat32:
            return hail.float32(self)
        else:
            assert typ == tfloat64
            return hail.float64(self)

    def _promote_numeric(self, typ):
        coercer = expressions.coercer_from_dtype(typ)
        if isinstance(typ, tarray) and not isinstance(self.dtype, tarray):
            return coercer.ec.coerce(self)
        elif isinstance(typ, tndarray) and not isinstance(self.dtype, tndarray):
            return coercer.ec.coerce(self)
        else:
            return coercer.coerce(self)

    @staticmethod
    def _div_ret_type_f(t):
        assert is_numeric(t)
        if t == tint32 or t == tint64:
            return tfloat32
        else:
            # Float64 or Float32
            return t

    def _bin_op_numeric_unify_types(self, name, other):
        def numeric_proxy(t):
            if t == tbool:
                return tint32
            else:
                return t

        def scalar_type(t):
            if isinstance(t, tarray):
                return numeric_proxy(t.element_type)
            elif isinstance(t, tndarray):
                return numeric_proxy(t.element_type)
            else:
                return numeric_proxy(t)

        t = unify_types(scalar_type(self.dtype), scalar_type(other.dtype))
        if t is None:
            raise NotImplementedError("'{}' {} '{}'".format(self.dtype, name, other.dtype))

        if isinstance(self.dtype, tarray) or isinstance(other.dtype, tarray):
            return tarray(t)
        elif isinstance(self.dtype, tndarray):
            return tndarray(t, self.ndim)
        elif isinstance(other.dtype, tndarray):
            return tndarray(t, other.ndim)

        return t

    def _bin_op_numeric(self, name, other, ret_type_f=None):
        other = to_expr(other)
        unified_type = self._bin_op_numeric_unify_types(name, other)
        me = self._promote_numeric(unified_type)
        other = other._promote_numeric(unified_type)
        if ret_type_f:
            if isinstance(unified_type, tarray):
                ret_type = tarray(ret_type_f(unified_type.element_type))
            elif isinstance(unified_type, tndarray):
                ret_type = tndarray(ret_type_f(unified_type.element_type), unified_type.ndim)
            else:
                ret_type = ret_type_f(unified_type)
        else:
            ret_type = unified_type
        return me._bin_op(name, other, ret_type)

    def _bin_op_numeric_reverse(self, name, other, ret_type_f=None):
        return to_expr(other)._bin_op_numeric(name, self, ret_type_f)

    def _unary_op(self, name):
        return expressions.construct_expr(ApplyUnaryPrimOp(name, self._ir), self._type, self._indices, self._aggregations)

    def _bin_op(self, name, other, ret_type):
        other = to_expr(other)
        indices, aggregations = unify_all(self, other)
        if (name in {'+', '-', '*', '/', '//'}) and (ret_type in {tint32, tint64, tfloat32, tfloat64}):
            op = ApplyBinaryPrimOp(name, self._ir, other._ir)
        elif name in {"==", "!=", "<", "<=", ">", ">="}:
            op = ApplyComparisonOp(name, self._ir, other._ir)
        else:
            d = {
                '+': 'add', '-': 'sub', '*': 'mul', '/': 'div', '//': 'floordiv',
                '%': 'mod', '**': 'pow'
            }
            op = Apply(d.get(name, name), ret_type, self._ir, other._ir)
        return expressions.construct_expr(op, ret_type, indices, aggregations)

    def _bin_op_reverse(self, name, other, ret_type):
        return to_expr(other)._bin_op(name, self, ret_type)

    def _method(self, name, ret_type, *args):
        args = tuple(to_expr(arg) for arg in args)
        indices, aggregations = unify_all(self, *args)
        ir = Apply(name, ret_type, self._ir, *(a._ir for a in args))
        return expressions.construct_expr(ir, ret_type, indices, aggregations)

    def _index(self, ret_type, key):
        key = to_expr(key)
        return self._method("index", ret_type, key)

    def _slice(self, ret_type, start=None, stop=None, step=None):
        if step is not None:
            raise NotImplementedError('Variable slice step size is not currently supported')

        if start is not None:
            start = to_expr(start)
            if stop is not None:
                stop = to_expr(stop)
                return self._method('slice', ret_type, start, stop)
            else:
                return self._method('sliceRight', ret_type, start)
        else:
            if stop is not None:
                stop = to_expr(stop)
                return self._method('sliceLeft', ret_type, stop)
            else:
                return self

    def _ir_lambda_method(self, irf, f, input_type, ret_type_f, *args):
        args = (to_expr(arg)._ir for arg in args)
        new_id = Env.get_uid()
        lambda_result = to_expr(
            f(expressions.construct_variable(new_id, input_type, self._indices, self._aggregations)))

        indices, aggregations = unify_all(self, lambda_result)
        ir = irf(self._ir, new_id, lambda_result._ir, *args)
        return expressions.construct_expr(ir, ret_type_f(lambda_result._type), indices, aggregations)

    @property
    def dtype(self) -> HailType:
        """The data type of the expression.

        Returns
        -------
        :class:`.HailType`

        """
        return self._type

    def __len__(self):
        raise TypeError("'Expression' objects have no static length: use 'hl.len' for the length of collections")

    def __hash__(self):
        return super(Expression, self).__hash__()

    def __repr__(self):
        return f'<{self.__class__.__name__} of type {self.dtype}>'

    def __eq__(self, other):
        """Returns ``True`` if the two expressions are equal.

        Examples
        --------

        >>> x = hl.literal(5)
        >>> y = hl.literal(5)
        >>> z = hl.literal(1)

        >>> hl.eval(x == y)
        True

        >>> hl.eval(x == z)
        False

        Notes
        -----
        This method will fail with an error if the two expressions are not
        of comparable types.

        Parameters
        ----------
        other : :class:`.Expression`
            Expression for equality comparison.

        Returns
        -------
        :class:`.BooleanExpression`
            ``True`` if the two expressions are equal.
        """
        return self._compare_op("==", other)

    def __ne__(self, other):
        """Returns ``True`` if the two expressions are not equal.

        Examples
        --------

        >>> x = hl.literal(5)
        >>> y = hl.literal(5)
        >>> z = hl.literal(1)

        >>> hl.eval(x != y)
        False

        >>> hl.eval(x != z)
        True

        Notes
        -----
        This method will fail with an error if the two expressions are not
        of comparable types.

        Parameters
        ----------
        other : :class:`.Expression`
            Expression for inequality comparison.

        Returns
        -------
        :class:`.BooleanExpression`
            ``True`` if the two expressions are not equal.
        """
        return self._compare_op("!=", other)

    def _to_table(self, name):
        name, ds = self._to_relational(name)
        if isinstance(ds, hail.MatrixTable):
            entries = ds.key_cols_by().entries()
            entries = entries.order_by(*ds.row_key)
            return name, entries.select(name)
        else:
            if len(ds.key) != 0:
                ds = ds.order_by(*ds.key)
            return name, ds.select(name)

    def _to_relational(self, fallback_name):
        source = self._indices.source
        axes = self._indices.axes
        if not self._aggregations.empty():
            raise NotImplementedError('cannot convert aggregated expression to table')

        if source is None:
            return fallback_name, hl.Table.parallelize([hl.struct(**{fallback_name: self})], n_partitions=1)

        name = source._fields_inverse.get(self)
        top_level = name is not None
        if not top_level:
            name = fallback_name
        named_self = {name: self}
        if len(axes) == 0:
            x = source.select_globals(**named_self)
            ds = hl.Table.parallelize([x.index_globals()], n_partitions=1)
        elif isinstance(source, hail.Table):
            if top_level and name in source.key:
                named_self = {}
            ds = source.select(**named_self).select_globals()
        elif isinstance(source, hail.MatrixTable):
            if self._indices == source._row_indices:
                if top_level and name in source.row_key:
                    named_self = {}
                ds = source.select_rows(**named_self).select_globals().rows()
            elif self._indices == source._col_indices:
                if top_level and name in source.col_key:
                    named_self = {}
                ds = source.select_cols(**named_self).select_globals().key_cols_by().cols()
            else:
                assert self._indices == source._entry_indices
                ds = source.select_entries(**named_self).select_globals().select_cols().select_rows()
        return name, ds

    @typecheck_method(n=nullable(int),
                      width=nullable(int),
                      truncate=nullable(int),
                      types=bool,
                      handler=nullable(anyfunc),
                      n_rows=nullable(int),
                      n_cols=nullable(int))
    def show(self,
             n=None,
             width=None,
             truncate=None,
             types=True,
             handler=None,
             n_rows=None,
             n_cols=None):
        """Print the first few rows of the table to the console.

        Examples
        --------

        >>> table1.SEX.show()
        +-------+-----+
        |    ID | SEX |
        +-------+-----+
        | int32 | str |
        +-------+-----+
        |     1 | "M" |
        |     2 | "M" |
        |     3 | "F" |
        |     4 | "F" |
        +-------+-----+

        >>> hl.literal(123).show()
        +--------+
        | <expr> |
        +--------+
        |  int32 |
        +--------+
        |    123 |
        +--------+

        Warning
        -------
        Extremely experimental.

        Parameters
        ----------
        n : :obj:`int`
            Maximum number of rows to show.
        width : :obj:`int`
            Horizontal width at which to break columns.
        truncate : :obj:`int`, optional
            Truncate each field to the given number of characters. If
            ``None``, truncate fields to the given `width`.
        types : :obj:`bool`
            Print an extra header line with the type of each field.
        """
        kwargs = {
            'n': n, 'width': width, 'truncate': truncate, 'types': types,
            'handler': handler, 'n_rows': n_rows, 'n_cols': n_cols}
        if kwargs.get('n_rows') is None:
            kwargs['n_rows'] = kwargs['n']
        del kwargs['n']
        _, ds = self._to_relational_preserving_rows_and_cols('<expr>')
        ds.show(**{k: v for k, v in kwargs.items() if v is not None})

    def _to_relational_preserving_rows_and_cols(self, fallback_name):
        source = self._indices.source
        if isinstance(source, hl.Table):
            if self is source.row:
                return None, source
            if self is source.key:
                return None, source.select()
        if isinstance(source, hl.MatrixTable):
            if self is source.row:
                return None, source.rows()
            if self is source.row_key:
                return None, source.rows().select()
            if self is source.col:
                return None, source.key_cols_by().cols()
            if self is source.col_key:
                return None, source.key_cols_by().cols().select()
            if self is source.entry:
                return None, source.select_rows().select_cols()
        return self._to_relational(fallback_name)

    @typecheck_method(path=str, delimiter=str, missing=str, header=bool)
    def export(self, path, delimiter='\t', missing='NA', header=True):
        """Export a field to a text file.

        Examples
        --------

        >>> small_mt.GT.export('output/gt.tsv')
        >>> with open('output/gt.tsv', 'r') as f:
        ...     for line in f:
        ...         print(line, end='')
        locus	alleles	0	1	2	3
        1:1	["A","C"]	0/1	0/1	0/0	0/0
        1:2	["A","C"]	1/1	0/1	1/1	1/1
        1:3	["A","C"]	1/1	0/1	0/1	0/0
        1:4	["A","C"]	1/1	0/1	1/1	1/1
        <BLANKLINE>

        >>> small_mt.GT.export('output/gt-no-header.tsv', header=False)
        >>> with open('output/gt-no-header.tsv', 'r') as f:
        ...     for line in f:
        ...         print(line, end='')
        1:1	["A","C"]	0/1	0/1	0/0	0/0
        1:2	["A","C"]	1/1	0/1	1/1	1/1
        1:3	["A","C"]	1/1	0/1	0/1	0/0
        1:4	["A","C"]	1/1	0/1	1/1	1/1
        <BLANKLINE>

        >>> small_mt.pop.export('output/pops.tsv')
        >>> with open('output/pops.tsv', 'r') as f:
        ...     for line in f:
        ...         print(line, end='')
        sample_idx	pop
        0	2
        1	2
        2	0
        3	2
        <BLANKLINE>

        >>> small_mt.ancestral_af.export('output/ancestral_af.tsv')
        >>> with open('output/ancestral_af.tsv', 'r') as f:
        ...     for line in f:
        ...         print(line, end='')
        locus	alleles	ancestral_af
        1:1	["A","C"]	5.3905e-01
        1:2	["A","C"]	8.6768e-01
        1:3	["A","C"]	4.3765e-01
        1:4	["A","C"]	7.6300e-01
        <BLANKLINE>

        >>> mt = small_mt
        >>> small_mt.bn.export('output/bn.tsv')
        >>> with open('output/bn.tsv', 'r') as f:
        ...     for line in f:
        ...         print(line, end='')
        bn
        {"n_populations":3,"n_samples":4,"n_variants":4,"n_partitions":8,"pop_dist":[1,1,1],"fst":[0.1,0.1,0.1],"mixture":false}
        <BLANKLINE>


        Notes
        -----

        For entry-indexed expressions, if there is one column key field, the
        result of calling :func:`hl.str` on that field is used as the column
        header. Otherwise, each compound column key is converted to JSON and
        used as a column header. For example:

        >>> small_mt = small_mt.key_cols_by(s=small_mt.sample_idx, family='fam1')
        >>> small_mt.GT.export('output/gt-no-header.tsv')
        >>> with open('output/gt-no-header.tsv', 'r') as f:
        ...     for line in f:
        ...         print(line, end='')
        locus	alleles	{"s":0,"family":"fam1"}	{"s":1,"family":"fam1"}	{"s":2,"family":"fam1"}	{"s":3,"family":"fam1"}
        1:1	["A","C"]	0/1	0/1	0/0	0/0
        1:2	["A","C"]	1/1	0/1	1/1	1/1
        1:3	["A","C"]	1/1	0/1	0/1	0/0
        1:4	["A","C"]	1/1	0/1	1/1	1/1
        <BLANKLINE>


        Parameters
        ----------
        path : :obj:`str`
            The path to which to export.
        delimiter : :obj:`str`
            The string for delimiting columns.
        missing : :obj:`str`
            The string to output for missing values.
        header : :obj:`bool`
            When ``True`` include a header line.
        """
        uid = Env.get_uid()
        self_name, ds = self._to_relational_preserving_rows_and_cols(uid)
        if isinstance(ds, hl.Table):
            ds.export(output=path, delimiter=delimiter, header=header)
        else:
            assert len(self._indices.axes) == 2
            entries, cols = Env.get_uid(), Env.get_uid()
            t = ds.select_cols().localize_entries(entries, cols)
            t = t.order_by(*t.key)
            output_col_name = Env.get_uid()
            entry_array = t[entries]
            if self_name:
                entry_array = hl.map(lambda x: x[self_name], entry_array)
            entry_array = hl.map(lambda x: hl.cond(hl.is_missing(x), missing, hl.str(x)),
                                 entry_array)
            file_contents = t.select(
                **{k: hl.str(t[k]) for k in ds.row_key},
                **{output_col_name: hl.delimit(entry_array, delimiter)})
            if header:
                col_key = t[cols]
                if len(ds.col_key) == 1:
                    col_key = hl.map(lambda x: x[0], col_key)
                column_names = hl.map(hl.str, col_key).collect(_localize=False)[0]
                header_table = hl.utils.range_table(1).key_by().select(
                    **{k: k for k in ds.row_key},
                    **{output_col_name: hl.delimit(column_names, delimiter)})
                file_contents = header_table.union(file_contents)
            file_contents.export(path, delimiter=delimiter, header=False)


    @typecheck_method(n=int, _localize=bool)
    def take(self, n, _localize=True):
        """Collect the first `n` records of an expression.

        Examples
        --------

        Take the first three rows:

        >>> table1.X.take(3)
        [5, 6, 7]

        Warning
        -------
        Extremely experimental.

        Parameters
        ----------
        n : int
            Number of records to take.

        Returns
        -------
        :obj:`list`
        """
        uid = Env.get_uid()
        name, t = self._to_table(uid)
        e = t.take(n, _localize=False).map(lambda r: r[name])
        if _localize:
            return hl.eval(e)
        return e

    @typecheck_method(_localize=bool)
    def collect(self, _localize=True):
        """Collect all records of an expression into a local list.

        Examples
        --------

        Collect all the values from `C1`:

        >>> table1.C1.collect()
        [2, 2, 10, 11]

        Warning
        -------
        Extremely experimental.

        Warning
        -------
        The list of records may be very large.

        Returns
        -------
        :obj:`list`
        """
        uid = Env.get_uid()
        name, t = self._to_table(uid)
        e = t.collect(_localize=False).map(lambda r: r[name])
        if _localize:
            return hl.eval(e)
        return e

    def _extra_summary_fields(self, agg_result):
        return {}

    def _summary_fields(self, agg_result, top):
        if top:
            return {}, self._nested_summary(agg_result[2], top)
        n_missing = agg_result[0]
        n_defined = agg_result[1]
        tot = n_missing + n_defined
        missing_value_str = str(n_missing) if n_missing == 0 else f'{n_missing} ({(n_missing / tot) * 100:.2f}%)'
        defined_value_str = str(n_defined) if n_defined == 0 else f'{n_defined} ({(n_defined / tot) * 100:.2f}%)'
        if n_defined == 0:
            return {'Non-missing': defined_value_str, 'Missing': missing_value_str}, {}
        return {'Non-missing': defined_value_str, 'Missing': missing_value_str,
                **self._extra_summary_fields(agg_result[2])}, self._nested_summary(agg_result[2], top)

    def _nested_summary(self, agg_result, top):
        return {}

    def _summary_aggs(self):
        return hl.null(hl.tint32)

    def _all_summary_aggs(self):
        return hl.tuple((
            hl.agg.filter(hl.is_missing(self), hl.agg.count()),
            hl.agg.filter(hl.is_defined(self), hl.agg.count()),
            self._summary_aggs()))

    def _summarize(self, agg_res=None, *, name=None, header=None, top=False):
        src = self._indices.source
        summary_header = None
        if src is None or len(self._indices.axes) == 0:
            raise ValueError("Cannot summarize a scalar expression")
        if agg_res is None:
            count, agg_res = self._aggregation_method()(hl.tuple((hl.agg.count(), self._all_summary_aggs())))
            summary_header = f'{count} records.'
        sum_fields, nested = self._summary_fields(agg_res, top)
        summary = Summary(self._type, agg_res[0], sum_fields, nested, header=summary_header)
        if name is None and header is None:
            return summary
        else:
            return NamedSummary(summary, name, header)

    def summarize(self, handler=None):
        """Compute and print summary information about the expression.

        .. include:: _templates/experimental.rst
        """

        src = self._indices.source
        if self in src._fields:
            field_name = src._fields_inverse[self]
            prefix = field_name
        else:
            if self._ir.is_nested_field:
                prefix = self._ir.name
            else:
                prefix = '<expr>'

        if handler is None:
            handler = hl.utils.default_handler()
        handler(self._summarize(name=prefix))

    def _selector_and_agg_method(self):
        src = self._indices.source
        assert src is not None
        assert len(self._indices.axes) > 0
        if isinstance(src, hl.MatrixTable):
            if self._indices == src._row_indices:
                return src.select_rows, lambda t: t.aggregate_rows
            elif self._indices == src._col_indices:
                return src.select_cols, lambda t: t.aggregate_cols
            else:
                return src.select_entries, lambda t: t.aggregate_entries
        else:
            return src.select, lambda t: t.aggregate

    def _aggregation_method(self):
        return self._selector_and_agg_method()[1](self._indices.source)



from .blockmatrix import BlockMatrix, _jarray_from_ndarray, _breeze_from_ndarray, _svd, _eigh
from . import utils as utils

__all__ = ['BlockMatrix',
           'utils',
           '_jarray_from_ndarray',
           '_breeze_from_ndarray',
           '_svd',
           '_eigh']
